package org.LiveGraph.settings;

import java.awt.Color;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Properties;

import org.LiveGraph.LiveGraph;

/**
 * Encapsulates settings concerned with the data graphs and the plot canvas.
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: GraphSettings.java</p> 
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p>
 * 
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class GraphSettings extends ObservableSettings {

/**
 * Standard file extension.
 */
public static final String preferredFileExtension = ".lggs";

/**
 * Possible types for the vertical grid.
 */
public static enum VGridType { VGrid_None, VGrid_DSNumAligned, VGrid_XAUnitAligned };

/**
 * Possible types for the horizontal grid.
 */
public static enum HGridType { HGrid_None, HGrid_Simple };

/**
 * Possible types for the x-axis.
 */
public static enum XAxisType { XAxis_DSNum, XAxis_DataValSimple, XAxis_DataValTrans0to1, XAxis_DataValScaledSet };

/**
 * Default settings.
 */
private static final Properties defaultValues = new Properties();
static {
	defaultValues.setProperty("MinY", "Auto");
	defaultValues.setProperty("MaxY", "Auto");
	defaultValues.setProperty("MinX", "Auto");
	defaultValues.setProperty("MaxX", "Auto");
	defaultValues.setProperty("VGridType", "VGrid_None");
	defaultValues.setProperty("VGridSize", "50");
	defaultValues.setProperty("VGridColour", "C0C0C0");
	defaultValues.setProperty("HGridType", "HGrid_None");
	defaultValues.setProperty("HGridSize", "50");
	defaultValues.setProperty("HGridColour", "C0C0C0");
	defaultValues.setProperty("XAxisType", "XAxis_DSNum");
	defaultValues.setProperty("XAxisSeriesIndex", "0");
	defaultValues.setProperty("XAxisScaleValue", "100");
	defaultValues.setProperty("HighlightDataPoints", "1");
}

/**
 * Holds the current graph settings.
 */
private Properties values = null;

/**
 * Creates a new graph settings object with the default settings.
 */
public GraphSettings() {
	super();
	values = new Properties();
	values.putAll(defaultValues);
	checkDisableHighlightingForOldJava();
}

/**
 * Creates a new graph settings object and loads the settings from the specified file.
 * 
 * @param fileName File to load the settigs from.
 */
public GraphSettings(String fileName) {
	this();
	load(fileName);
}

/**
 * Loads the settings from the specified file.
 * 
 * @param fileName File to load the settigs from.
 * @return {@code true} if the settings were loaded, {@code false} if an exception occured.
 */
public boolean load(String fileName) {
	try {
		FileInputStream in = new FileInputStream(fileName);
		try {
			values.loadFromXML(in);
			
			if (values.containsKey("SavedWithIncompatibleJavaVersion")
					&& "1".equals(values.getProperty("SavedWithIncompatibleJavaVersion"))
					&& !getHighlightDataPoints()
					&& LiveGraph.application().runsCorrectJavaVersion()) {
				values.setProperty("HighlightDataPoints", "1");
			}
						
			checkDisableHighlightingForOldJava();
			notifyObservers("load");
		} finally { in.close(); }		
		return true;
	} catch(IOException e) {
		e.printStackTrace();
		return false;
	}
}

/**
 * Saves the settings to a specified file.
 * 
 * @param fileName The file to save the settings to.
 * @return {@code true} if the settings were saved, {@code false} if an exception occured. 
 */
public boolean save(String fileName) {
	
	if(LiveGraph.application().runsCorrectJavaVersion())
		values.remove("SavedWithIncompatibleJavaVersion");
	else
		values.setProperty("SavedWithIncompatibleJavaVersion", "1");
	
	try {
		FileOutputStream out = new FileOutputStream(fileName);
		try { values.storeToXML(out, "LiveGraph version " + LiveGraph.version + ". GraphSettings."); }
		finally { out.close(); }
		return true;
	} catch(IOException e) {
		e.printStackTrace();
		return false;
	}
}

/**
 * Gets the minimum Y value for the plot viewport.
 * 
 * @return The minimum value along the Y axis for the graph view
 * or {@code java.lang.Double.NaN} if the current global minimum of all data series
 * should be used instead.  
 */
public double getMinY() {
	if ("Auto".equalsIgnoreCase(values.getProperty("MinY")))
		return Double.NaN;
	else
		return Double.parseDouble(values.getProperty("MinY"));
}

/**
 * Sets the minimum Y value for the plot viewport.
 * 
 * @param v The minimum value along the Y axis for the graph view
 * or {@code java.lang.Double.NaN} if the global minimum of all data series
 * should be used at all times.  
 */
public void setMinY(double v) {
	if (Double.isNaN(v))
		values.setProperty("MinY", "Auto");
	else
		values.setProperty("MinY", Double.toString(v));	
	notifyObservers("MinY");
}

/**
 * Gets the maximum Y value for the plot viewport.
 * 
 * @return The maximum value along the Y axis for the graph view
 * or {@code java.lang.Double.NaN} if the current global maximum of all data series
 * should be used instead.  
 */
public double getMaxY() {
	if ("Auto".equalsIgnoreCase(values.getProperty("MaxY")))
		return Double.NaN;
	else
		return Double.parseDouble(values.getProperty("MaxY"));
}

/**
 * Sets the maximum Y value for the plot viewport.
 * 
 * @param v The maximum value along the Y axis for the graph view
 * or {@code java.lang.Double.NaN} if the global maximum of all data series
 * should be used at all times.  
 */
public void setMaxY(double v) {
	if (Double.isNaN(v))
		values.setProperty("MaxY", "Auto");
	else
		values.setProperty("MaxY", Double.toString(v));
	notifyObservers("MaxY");
}

/**
 * Gets the minimum X value for the plot viewport.
 * 
 * @return The minimum value along the X axis for the graph view
 * or {@code java.lang.Double.NaN} if the currently smallest X value of all data series
 * should be used instead.  
 */
public double getMinX() {
	if ("Auto".equalsIgnoreCase(values.getProperty("MinX")))
		return Double.NaN;
	else
		return Double.parseDouble(values.getProperty("MinX"));
}

/**
 * Sets the minimum X value for the plot viewport.
 * 
 * @param v The minimum value along the X axis for the graph view
 * or {@code java.lang.Double.NaN} if the smallest X value of all data series
 * should be used at all times.  
 */
public void setMinX(double v) {
	if (Double.isNaN(v))
		values.setProperty("MinX", "Auto");
	else
		values.setProperty("MinX", Double.toString(v));
	notifyObservers("MinX");
}

/**
 * Gets the maximum X value for the plot viewport.
 * 
 * @return The maximum value along the X axis for the graph view
 * or {@code java.lang.Double.NaN} if the currently largest X value of all data series
 * should be used instead.  
 */
public double getMaxX() {
	if ("Auto".equalsIgnoreCase(values.getProperty("MaxX")))
		return Double.NaN;
	else
		return Double.parseDouble(values.getProperty("MaxX"));
}

/**
 * Sets the maximum X value for the plot viewport.
 * 
 * @param v The maximum value along the X axis for the graph view
 * or {@code java.lang.Double.NaN} if the largest X value of all data series
 * should be used at all times.  
 */
public void setMaxX(double v) {
	if (Double.isNaN(v))
		values.setProperty("MaxX", "Auto");
	else
		values.setProperty("MaxX", Double.toString(v));
	notifyObservers("MaxX");
}

/**
 * Sets the vertical grid type.
 * 
 * @return The vertical grid type.
 */
public VGridType getVGridType() {
	if ("VGrid_DSNumAligned".equalsIgnoreCase(values.getProperty("VGridType")))
		return VGridType.VGrid_DSNumAligned;
	if ("VGrid_XAUnitAligned".equalsIgnoreCase(values.getProperty("VGridType")))
		return VGridType.VGrid_XAUnitAligned;
	return VGridType.VGrid_None;
}

/**
 * Gets the vertical grid type.
 * 
 * @param v The vertical grid type.
 */
public void setVGridType(VGridType v) {
	switch(v) {
		case VGrid_DSNumAligned:	values.setProperty("VGridType", "VGrid_DSNumAligned"); break;
		case VGrid_XAUnitAligned:	values.setProperty("VGridType", "VGrid_XAUnitAligned"); break;
		default:					values.setProperty("VGridType", "VGrid_None"); break;
	}
	notifyObservers("VGridType");
}

/**
 * Gets the interval between the vertical grid bars.
 * 
 * @return The interval between the vertical grid bars.
 */
public double getVGridSize() {
	return Double.parseDouble(values.getProperty("VGridSize"));
}

/**
 * Sets the interval between the vertical grid bars.
 * 
 * @param val The interval between the vertical grid bars.
 */
public void setVGridSize(double val) {
	if (Double.isNaN(val) || Double.isInfinite(val) || 0 > val)
		val = 0;	
	values.setProperty("VGridSize", Double.toString(val));
	notifyObservers("VGridSize");
}

/**
 * Gets the colour to use for drawing the vertical grid bars.
 * 
 * @return The colour to use for drawing the vertical grid bars.
 */
public Color getVGridColour() {
	String c = values.getProperty("VGridColour");
	int r, g, b;
	try { r = Integer.parseInt(c.substring(0, 2), 16); } catch(NumberFormatException e) { r = 0; }
	try { g = Integer.parseInt(c.substring(2, 4), 16); } catch(NumberFormatException e) { g = 0; }
	try { b = Integer.parseInt(c.substring(4, 6), 16); } catch(NumberFormatException e) { b = 0; }
	return new Color(r, g, b);
}

/**
 * Sets the colour to use for drawing the vertical grid bars.
 * @param c The colour to use for drawing the vertical grid bars.
 */
public void setVGridColour(Color c) {	
	if (null == c)
		values.setProperty("VGridColour", "000000");
	else
		values.setProperty("VGridColour", String.format("%02x%02x%02x", c.getRed(), c.getGreen(), c.getBlue()));
	notifyObservers("VGridColour");
}

/**
 * Gets the horizontal grid type.
 * @return The horizontal grid type.
 */
public HGridType getHGridType() {	
	if ("HGrid_Simple".equalsIgnoreCase(values.getProperty("HGridType")))
		return HGridType.HGrid_Simple;
	return HGridType.HGrid_None;
}

/**
 * Sets the horizontal grid type.
 * @param v The horizontal grid type.
 */
public void setHGridType(HGridType v) {
	switch(v) {
		case HGrid_Simple:	values.setProperty("HGridType", "HGrid_Simple"); break;
		default:			values.setProperty("HGridType", "HGrid_None"); break;
	}
	notifyObservers("HGridType");
}

/**
 * Gets the interval between the horizontal grib bars.
 * 
 * @return The interval between the horizontal grib bars.
 */
public double getHGridSize() {
	return Double.parseDouble(values.getProperty("HGridSize"));
}

/**
 * Sets the interval between the horizontal grib bars.
 * @param val The interval between the horizontal grib bars.
 */
public void setHGridSize(double val) {	
	if (Double.isNaN(val) || Double.isInfinite(val) || 0 > val)
		val = 0;	
	values.setProperty("HGridSize", Double.toString(val));
	notifyObservers("HGridSize");	
}


/**
 * Gets the colour for drawing the horizontal grid bars.
 * 
 * @return The colour for drawing the horizontal grid bars.
 */
public Color getHGridColour() {
	String c = values.getProperty("HGridColour");
	int r, g, b;
	try { r = Integer.parseInt(c.substring(0, 2), 16); } catch(NumberFormatException e) { r = 0; }
	try { g = Integer.parseInt(c.substring(2, 4), 16); } catch(NumberFormatException e) { g = 0; }
	try { b = Integer.parseInt(c.substring(4, 6), 16); } catch(NumberFormatException e) { b = 0; }
	return new Color(r, g, b);
}

/**
 * Sets the colour for drawing the horizontal grid bars.
 * 
 * @param c The colour for drawing the horizontal grid bars.
 */
public void setHGridColour(Color c) {	
	if (null == c)
		values.setProperty("HGridColour", "000000");
	else
		values.setProperty("HGridColour", String.format("%02x%02x%02x", c.getRed(), c.getGreen(), c.getBlue()));
	notifyObservers("HGridColour");
}

/**
 * Gets the type for the x-axis.
 * 
 * @return The type for the x-axis.
 */
public XAxisType getXAxisType() {	
	if ("XAxis_DataValSimple".equalsIgnoreCase(values.getProperty("XAxisType")))
		return XAxisType.XAxis_DataValSimple;
	if ("XAxis_DataValTrans0to1".equalsIgnoreCase(values.getProperty("XAxisType")))
		return XAxisType.XAxis_DataValTrans0to1;
	if ("XAxis_DataValScaledSet".equalsIgnoreCase(values.getProperty("XAxisType")))
		return XAxisType.XAxis_DataValScaledSet;
	return XAxisType.XAxis_DSNum;

}

/**
 * Sets the type for the x-axis.
 * @param v The type for the x-axis.
 */
public void setXAxisType(XAxisType v) {
	switch(v) {
		case XAxis_DataValSimple:	 values.setProperty("XAxisType", "XAxis_DataValSimple"); break;
		case XAxis_DataValTrans0to1: values.setProperty("XAxisType", "XAxis_DataValTrans0to1"); break;
		case XAxis_DataValScaledSet: values.setProperty("XAxisType", "XAxis_DataValScaledSet"); break;
		default:					 values.setProperty("XAxisType", "XAxis_DSNum"); break;
	}
	notifyObservers("XAxisType");
}

/**
 * Gets the index of the data series to use as the x-axis.
 * 
 * @return The index of the data series to use as the x-axis.
 */
public int getXAxisSeriesIndex() {
	return Integer.parseInt(values.getProperty("XAxisSeriesIndex"));
}

/**
 * Sets the index of the data series to use as the x-axis.
 * 
 * @param val The index of the data series to use as the x-axis.
 */
public void setXAxisSeriesIndex(int val) {	
	if (0 > val)
		val = 0;	
	values.setProperty("XAxisSeriesIndex", Integer.toString(val));
	notifyObservers("XAxisSeriesIndex");
}

/**
 * Gets the scale parameter to use with the type {@code XAxis_DataValScaledSet} for
 * scaling of X values.
 * 
 * @return The scale parameter to use with the type {@code XAxis_DataValScaledSet} for
 * scaling of X values.
 */
public double getXAxisScaleValue() {
	double val = Double.parseDouble(values.getProperty("XAxisScaleValue"));
	if (Double.isNaN(val) || Double.isInfinite(val))
		return 1.;
	return val;
}

/**
 * Sets the scale parameter to use with the type {@code XAxis_DataValScaledSet} for
 * scaling of X values.
 * 
 * @param val scale parameter to use with the type {@code XAxis_DataValScaledSet} for
 * scaling of X values.
 */
public void setXAxisScaleValue(double val) {
	if (Double.isNaN(val) || Double.isInfinite(val))
		val = 1.;	
	values.setProperty("XAxisScaleValue", Double.toString(val));
	notifyObservers("XAxisScaleValue");
}

/**
 * Gets whether the data points in the vicinity of the mouse cursor should be highlighted.
 * (Note, this is a hidden setting and cannot be changes via the GUI.)
 * 
 * @return {@code true} if the data points in the vicinity of the mouse cursor should be highlighted,
 * {@code false} otherwise.
 */
public boolean getHighlightDataPoints() { return "1".equals(values.getProperty("HighlightDataPoints")); }

/**
 * Gets whether the data points in the vicinity of the mouse cursor should be highlighted.
 * (Note, this is a hidden setting and cannot be changes via the GUI.)
 * 
 * @param v {@code true} if the data points in the vicinity of the mouse cursor are to be highlighted,
 * {@code false} otherwise.
 */
public void setHighlightDataPoints(boolean v) {
	values.setProperty("HighlightDataPoints", v ? "1" : "0");
	checkDisableHighlightingForOldJava();
	notifyObservers("HighlightDataPoints");
}

/**
 * If the Java version is too old, data point highlighting is disabled.
 */
private void checkDisableHighlightingForOldJava() {
	if (!LiveGraph.application().runsCorrectJavaVersion())
		values.setProperty("HighlightDataPoints", "0");
}

} // public class GraphSettings